from contextlib import nullcontext
from contextlib import redirect_stdout
import json
import os
from pathlib import Path
import sys
import typing as t

from elasticsearch.helpers import scan
from elasticsearch.helpers import streaming_bulk
import typer
from uvicorn.importer import import_from_string

from . import es
from .handlers import ActionHandler
from .utils import jobj_to_line
from .utils import merge_dicts
from .utils import parse_header
from .utils import parse_params


app = typer.Typer()


@app.command(name='e')
def search(
    host: t.Annotated[str, typer.Argument(help='Elasticsearch host')] = 'localhost',
):
    client = es.Client(host=host)
    client.search


@app.command(name='s')
def scan_(
    verbose: t.Annotated[bool, typer.Option(' /-v', ' /--verbose')] = False,
    progress: t.Annotated[bool, typer.Option(' /-S', ' /--no-progress')] = True,
    file_input_search_body: t.Annotated[typer.FileText, typer.Option('-d', '--data')] = t.cast(
        typer.FileText, sys.stdin
    ),
    file_output: t.Annotated[typer.FileTextWrite, typer.Option('-o', '--output')] = t.cast(
        typer.FileTextWrite, sys.stdout
    ),
    host: t.Annotated[str, typer.Argument(help='Elasticsearch host')] = 'localhost',
    scroll: t.Annotated[str, typer.Option('--scroll', help='Scroll duration')] = '5m',
    raise_on_error: t.Annotated[bool, typer.Option(' /--no-raise-on-error')] = True,
    preserve_order: t.Annotated[bool, typer.Option('--preserve-order')] = False,
    size: t.Annotated[int, typer.Option('--size')] = 1000,
    request_timeout: t.Annotated[t.Optional[int], typer.Option('--request-timeout')] = None,
    clear_scroll: t.Annotated[bool, typer.Option(' /--keep-scroll')] = True,
):
    search_body = file_input_search_body.read()
    with redirect_stdout(sys.stderr if verbose else Path(os.devnull).open('w')):
        print('search_body', search_body)
    search_body = json.loads(search_body)
    client = es.Client(host=host)
    iterable = scan(
        client=client,
        query=search_body,
        scroll=scroll,
        raise_on_error=raise_on_error,
        preserve_order=preserve_order,
        size=size,
        request_timeout=request_timeout,
        clear_scroll=clear_scroll,
    )
    progressbar = nullcontext(iterable)
    if progress:
        progressbar = typer.progressbar(
            iterable=iterable, label='scan', show_pos=True, file=sys.stderr
        )
    with progressbar as iterable:
        for hit in iterable:
            file_output.write(jobj_to_line(hit))


@app.command(name='q')
def perform_request(
    file_output: t.Annotated[typer.FileTextWrite, typer.Option('-o', '--output')] = t.cast(
        typer.FileTextWrite, sys.stdout
    ),
    host: t.Annotated[str, typer.Argument(help='Elasticsearch host')] = 'localhost',
    method: t.Annotated[
        str, typer.Option('-X', '--method', '--request', help='HTTP method')
    ] = 'GET',
    path: t.Annotated[str, typer.Option('-P', '--path', help='HTTP path')] = '/',
    _headers: t.Annotated[
        t.Optional[list[dict]],
        typer.Option('-H', '--header', parser=parse_header, help='HTTP headers'),
    ] = None,
    _params: t.Annotated[
        t.Optional[list[dict]],
        typer.Option('-p', '--params', parser=parse_params, help='HTTP params'),
    ] = None,
):
    if 0:
        with redirect_stdout(sys.stderr):
            print(f'{host = }')
            print(f'{method = }')
    client = es.Client(host=host)

    if not path.startswith('/'):
        path = '/' + path
    if 0:
        with redirect_stdout(sys.stderr):
            print(f'{path = }')

    headers = merge_dicts(_headers)
    params = merge_dicts(_params)
    if 0:
        with redirect_stdout(sys.stderr):
            print(f'{headers = }')
            print(f'{params = }')

    if 1:
        response = client.transport.perform_request(
            method=method,
            url=path,
            headers=headers,
            params=params,
        )
        file_output.write(jobj_to_line(response))


@app.command(name='t')
def streaming_bulk_(
    progress: t.Annotated[bool, typer.Option(' /-S', ' /--no-progress')] = True,
    file_input_bulk_body: t.Annotated[typer.FileText, typer.Option('-d', '--data')] = t.cast(
        typer.FileText, sys.stdin
    ),
    file_output: t.Annotated[typer.FileTextWrite, typer.Option('-o', '--output')] = t.cast(
        typer.FileTextWrite, sys.stdout
    ),
    host: t.Annotated[str, typer.Argument(help='Elasticsearch host')] = 'localhost',
    handler: t.Annotated[
        t.Callable[[t.Iterable[str]], t.Iterable[str]],
        typer.Option(
            '-H',
            '--handler',
            parser=import_from_string,
            help='Handler class. e.g. esqt.handlers.BaseActionsHandler',
        ),
    ] = ActionHandler,
    chunk_size: t.Annotated[int, typer.Option('--chunk-size')] = 500,
    max_chunk_bytes: t.Annotated[int, typer.Option('--max-chunk-bytes')] = 100 * 1024 * 1024,
    raise_on_error: t.Annotated[bool, typer.Option(' /--no-raise-on-error')] = True,
    raise_on_exception: t.Annotated[bool, typer.Option(' /--no-raise-on-exception')] = True,
    max_retries: t.Annotated[int, typer.Option('--max-retries')] = 3,
    initial_backoff: t.Annotated[int, typer.Option('--initial-backoff')] = 2,
    max_backoff: t.Annotated[int, typer.Option('--max-backoff')] = 600,
):
    client = es.Client(host=host)
    actions = handler(file_input_bulk_body)
    iterable = streaming_bulk(
        client=client,
        actions=actions,
        chunk_size=chunk_size,
        max_chunk_bytes=max_chunk_bytes,
        raise_on_error=raise_on_error,
        raise_on_exception=raise_on_exception,
        max_retries=max_retries,
        initial_backoff=initial_backoff,
        max_backoff=max_backoff,
        yield_ok=True,
    )
    progressbar = nullcontext(iterable)
    if progress:
        progressbar = typer.progressbar(
            iterable=iterable, label='streaming_bulk', show_pos=True, file=sys.stderr
        )
    with progressbar as iterable:
        success, failed = 0, 0
        for ok, item in iterable:
            if not ok:
                with redirect_stdout(sys.stderr):
                    print(f'Failed to index {item}')
                file_output.write(jobj_to_line(item))
                failed += 1
            else:
                with redirect_stdout(sys.stderr):
                    print('.', end='')
                file_output.write(jobj_to_line(item))
                success += 1
    with redirect_stdout(sys.stderr):
        print()
        print(f'{success = }')
        print(f'{failed = }')


def main():
    app()


if __name__ == '__main__':
    main()
